/*	$Id: connect.c 2007/07/05 19:11:26 mallet $ */

/* 
 * Copyright (c) 2007 Anthony Mallet                  --  Fri Jun 15 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
static const char id[] = "$Id: connect.c 2007/07/05 19:11:26 mallet $";

#include <err.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "shring.h"


/* --- netshring_connect --------------------------------------------- */

netshring_id_t
netshring_connect(const char *name, netshring_opt_t sync)
{
  netshring_id_t shid;
  struct addrinfo *hinfo;
  char *host, *port;
  char handshake[sizeof(SHRNG_MAGIC_SYNC)];
  int s;

  signal(SIGPIPE, SIG_IGN);

  /* allocate handler */
  shid = malloc(sizeof(*shid)); if (!shid) return NULL;
  shid->type = SHRNG_DRAIN;
  shid->clients = 0;
  shid->clientfd = NULL;
  shid->sockfd = socket(AF_INET, SOCK_STREAM, 0);
  if (shid->sockfd == -1) goto error;
  pthread_mutex_init(&shid->mutex, NULL);

  /* compute host and port */
  host = strdup(name); if (!host) goto error;
  port = strchr(host, ':');
  if (!port) {
    port = host;
    host = strdup("localhost");
  } else {
    *port++ = 0;
    port = strdup(port);
  }

  /* connect to source */
  if (getaddrinfo(host, NULL, NULL, &hinfo)) {
    perror(host); goto error;
  }
  if (hinfo->ai_family == AF_INET)
    ((struct sockaddr_in *)(hinfo->ai_addr))->sin_port =
      htons(netshring_hashstring(port) + 10000);
  else {
    perror("not AF_INET"); goto error;
  }
  s = connect(shid->sockfd, hinfo->ai_addr, hinfo->ai_addrlen);

  freeaddrinfo(hinfo);
  free(host);
  free(port);

  if (s < 0) { perror("connect"); goto error; }

  /* handshake the server */
  s = write(shid->sockfd,
	    sync == SHRNG_SYNC ? SHRNG_MAGIC_SYNC : SHRNG_MAGIC_ASYNC,
	    sizeof(SHRNG_MAGIC_SYNC));
  if (s < 0) { perror("handshake"); goto error; }
  if (s != sizeof(SHRNG_MAGIC_SYNC)) {
    perror("short handshake"); goto error;
  }

  if(sync == SHRNG_SYNC) {
    /* pop the server's answer */
    s = netshring_pop(shid, handshake, sizeof(handshake), SHRNG_SYNC);
    if (s < 0) { perror("sync handshake"); goto error; }
    if (s != sizeof(handshake)) {
      perror("short sync handshake"); goto error;
    }
    if (strcmp(handshake, SHRNG_MAGIC_SYNC)) {
      warnx("bad handshake, got `%s'", handshake);
      goto error;
    }

    warnx("drain handshake `%s'", handshake);
  }

  return shid;

error:
  if (shid) {
    if (shid->sockfd >= 0) close(shid->sockfd);
    free(shid);
  }
  return NULL;
}


/* --- netshring_disconnect ------------------------------------------ */

void
netshring_disconnect(netshring_id_t shid)
{
  if (!shid) return;
  if (shid->sockfd >= 0) close(shid->sockfd);
  free(shid);
}
