/*	$Id: create.c 2007/07/09 16:32:37 mallet $ */

/*
 * Copyright (c) 2007 Anthony Mallet                  --  Fri Jun 15 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
static const char id[] = "$Id: create.c 2007/07/09 16:32:37 mallet $";

#include <err.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "shring.h"


/* --- local data ---------------------------------------------------- */

static void *netshring_server(void *data);


/* --- netshring_create ---------------------------------------------- */

netshring_id_t
netshring_create(const char *name)
{
  netshring_id_t shid;
  struct linger linger;
  struct sockaddr_in addr;
  pthread_t server;
  int s, reuse;

  signal(SIGPIPE, SIG_IGN);

  /* allocate handler */
  shid = malloc(sizeof(*shid)); if (!shid) {
    warnx("no memory"); return NULL;
  }
  shid->type = SHRNG_SOURCE;
  shid->clients = 0;
  shid->clientfd = NULL;
  pthread_mutex_init(&shid->mutex, NULL);

  /* create server socket */
  shid->sockfd = socket(AF_INET, SOCK_STREAM, 0);
  if (shid->sockfd == -1) { warnx("no socket"); goto error; }

  reuse = 1;
  linger.l_onoff = 1;
  linger.l_linger = 0;
  if (setsockopt(shid->sockfd, SOL_SOCKET, SO_REUSEADDR,
                 (char *)&reuse, sizeof(reuse)) < 0) {
    warnx("can't set reuse option"); goto error;
  }
  if (setsockopt(shid->sockfd, SOL_SOCKET, SO_LINGER,
                 (char *)&linger, sizeof(linger)) < 0) {
    warnx("can't set no linger option"); goto error;
  }

  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons(netshring_hashstring(name) + 10000);
  warnx("listening on port %d", netshring_hashstring(name) + 10000);

  if (bind(shid->sockfd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in))) {
    warnx("can't bind"); goto error;
  }

  if (listen(shid->sockfd, 2) < 0) { warnx("can't listen"); goto error; }

  /* create server thread */
  s = pthread_create(&server, NULL, netshring_server, shid);

  return shid;

error:
  if (shid) {
    if (shid->sockfd >= 0) close(shid->sockfd);
    free(shid);
  }
  return NULL;
}


/* --- netshring_delete ---------------------------------------------- */

void
netshring_delete(netshring_id_t shid)
{
  if (!shid) return;

  pthread_mutex_lock(&shid->mutex);
  shid->type = SHRNG_CLOSED;
  pthread_mutex_unlock(&shid->mutex);
}


/* --- netshring_server ---------------------------------------------- */

static void *
netshring_server(void *data)
{
  netshring_id_t shid = data;
  char handshake[sizeof(SHRNG_MAGIC_SYNC)];
  int *clients;
  ssize_t s;
  int i, fd;

  pthread_mutex_lock(&shid->mutex);
  while(shid->type == SHRNG_SOURCE) {
    pthread_mutex_unlock(&shid->mutex);
    fd = accept(shid->sockfd, NULL, NULL);
    pthread_mutex_lock(&shid->mutex);
    if (fd < 0) continue;

    warnx("accepted connection");

    /* handshake */
    do
      s = read(fd, handshake, sizeof(handshake));
    while(s < 0 && errno == EINTR);
    if (s < 0) { perror("handshake"); close(fd); continue; }
    if (s != sizeof(handshake)) {
      perror("short handshake"); close(fd); continue;
    }
    if (strcmp(handshake, SHRNG_MAGIC_SYNC) &&
	strcmp(handshake, SHRNG_MAGIC_ASYNC)) {
      perror("wrong handshake"); close(fd); continue;
    }
    warnx("source handshake `%s'", handshake);

    /* push handshake answer */
    if (!strcmp(handshake, SHRNG_MAGIC_SYNC)) {
      netshring_blkhead_t h;
      struct iovec v[2];

      h.length = sizeof(SHRNG_MAGIC_SYNC);
      v[0].iov_base = &h;
      v[0].iov_len = sizeof(h);
      v[1].iov_base = SHRNG_MAGIC_SYNC;
      v[1].iov_len = sizeof(SHRNG_MAGIC_SYNC);

      s = writev(fd, v, 2);
      if (s < 0) { perror("writev"); close(fd); continue; }
      if (s != sizeof(h) + sizeof(SHRNG_MAGIC_SYNC)) {
	perror("short writev"); close(fd);  continue;
      }

      warnx("source replied `%s'", SHRNG_MAGIC_SYNC);
    }

    clients = realloc(shid->clientfd, (shid->clients+1) * sizeof(int));
    if (!clients) continue;

    clients[shid->clients] = fd;
    shid->clientfd = clients;
    shid->clients++;
  }

  for(i = 0; i < shid->clients; i++)
    close(shid->clientfd[i]);
  close(shid->sockfd);
  pthread_mutex_unlock(&shid->mutex);
  free(shid);

  return NULL;
}
