/*	$Id: read.c 2007/07/09 19:53:17 mallet $ */

/* 
 * Copyright (c) 2007 Anthony Mallet                  --  Fri Jun 15 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
static const char id[] = "$Id: read.c 2007/07/09 19:53:17 mallet $";

#include <err.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/uio.h>

#include "shring.h"


/* --- netshring_pop ------------------------------------------------- */

int
netshring_pop(netshring_id_t shid, void *buffer, size_t max, netshring_opt_t sync)
{
  long flags;
  ssize_t s, l, m;
  netshring_blkhead_t h;
  char *p;
  if (!shid || shid->type != SHRNG_DRAIN) return -SHRNG_INVAL;

  if (sync == SHRNG_ASYNC) {
    flags = fcntl(shid->sockfd, F_GETFL);
    if (flags < 0) { perror("setting async flag"); return -SHRNG_INVAL; }
    flags |= O_NONBLOCK;
    flags = fcntl(shid->sockfd, F_SETFL, flags);
    if (flags < 0) { perror("setting async flag"); return -SHRNG_INVAL; }
  }

  /* read header */
  p = (char *)&h;
  m = sizeof(h);
  for(l = 0; l < sizeof(h);) {
    s = read(shid->sockfd, p, m);
    if (s < 0 && errno == EINTR) continue;
    if (s < 0 && errno == EAGAIN) return 0;
    if (s < 0) { perror("read header"); return -SHRNG_INVAL; }
    l += s;
    m -= s;
    p += s;
  }

  if (sync == SHRNG_ASYNC) {
    flags &= ~O_NONBLOCK;
    flags = fcntl(shid->sockfd, F_SETFL, flags);
    if (flags < 0) { perror("setting async flag"); return -SHRNG_INVAL; }
  }

  /* read body */
  p = buffer;
  m = h.length;
  for(l = 0; l <h.length;) {
    s = read(shid->sockfd, p, m > max ? max : m);
    if (s < 0 && errno == EINTR) continue;
    if (s < 0) { perror("read body"); return -SHRNG_INVAL; }
    l += s;
    m -= s;
    p += s; if (p-(char *)buffer >= max) p = buffer;
  }

  if (h.length > max) return -SHRNG_INVAL;
  return h.length;
}


/* --- netshring_push ------------------------------------------------ */

int
netshring_push(netshring_id_t shid, const void *buffer, size_t length)
{
  ssize_t s, act;
  netshring_blkhead_t h;
  struct iovec v[2];
  int i, j;
  if (!shid || shid->type != SHRNG_SOURCE) return -SHRNG_INVAL;

  /* setup header */
  h.length = length;
  v[0].iov_base = &h;
  v[0].iov_len = sizeof(h);

  /* setup body */
  v[1].iov_base = (void *)buffer;
  v[1].iov_len = length;

  /* send XXX should handle errors more cleanly */
  pthread_mutex_lock(&shid->mutex);
  if (shid->clients == 0) {
    pthread_mutex_unlock(&shid->mutex);
    warnx("dropped %d bytes as there is not listener", length);
    return 0;
  }

  act = 0;
  for(i = 0; i<shid->clients; i++) {
    do {
      s = writev(shid->clientfd[i], v, 2);
    } while(s < 0 && errno == EINTR);

    if (s < 0) {
      perror("writev");
      close(shid->clientfd[i]);
      for(j = i; j < shid->clients - 1; j++)
	shid->clientfd[j] = shid->clientfd[j+1];
      shid->clients--;
      i--;
      continue;
    }
    if (s != sizeof(h) + length) { perror("short writev"); continue; }
    act = length;
  }
  pthread_mutex_unlock(&shid->mutex);

  return act;
}
